# -*- coding: utf-8 -*-

# Constants as specified in http://data.rcc-acis.org/doc/


# http://data.rcc-acis.org/doc/#title2
AUTHORITIES = [
    ('1', "WBAN", "Weather Bureau Army Navy"),
    ('2', "COOP", "Cooperative Observer Network"),
    ('3', "FAA", "Federal Aviation Administration"),
    ('4', "WMO", "World Meteorological Organization"),
    ('5', "ICAO", "International Civil Aviation Organization"),
    ('6', "GHCN", "Global Historical Climate Network"),
    ('7', "NWSLI", "National Weather Service Location Identifier"),
    ('8', "RCC", "Regional Climate Centers"),
    ('9', "ThreadEx", "NOAA ThreadEx"),
    ('10', "CoCoRaHS", "Community Collaborative Rain, Hail and Snow Network"),
]

AUTHORITY_BY_ID = {
    id: {
        'id': id,
        'name': name,
        'desc': desc
    } for id, name, desc in AUTHORITIES
}

AUTHORITY_BY_NAME = {
    auth['name']: auth
    for auth in AUTHORITY_BY_ID.values()
}

# http://data.rcc-acis.org/doc/#title5
ELEMENTS = (
    ('1', "maxt", "Maximum temperature (°F)"),
    ('2', "mint", "Minimum temperature (°F)"),
    ('43', "avgt", "Average temperature (°F)"),
    ('3', "obst", "Obs time temperature (°F)"),
    ('4', "pcpn", "Precipitation (inches)"),
    ('10', "snow", "Snowfall (inches)"),
    ('11', "snwd", "Snow depth (inches)"),
    ('7', None, "Pan evap (inches)"),
    ('44', "cdd", "Degree days above base (CDD) (default base 65)"),
    ('45', "hdd", "Degree days below base (HDD) (default base 65)"),
    (None, "gdd", "Growing Degree Days (base 50)")
)

ELEMENT_BY_ID = {
    id: {
        'id': id,
        'name': name,
        'desc': desc
    } for id, name, desc in ELEMENTS if id is not None
}

ELEMENT_BY_NAME = {
    name: {
        'id': id,
        'name': name,
        'desc': desc
    } for id, name, desc in ELEMENTS if name is not None
}



import json
from datetime import datetime, date, timedelta
from wq.io import JsonNetIO, BaseIO, TimeSeriesMapper, make_date_mapper
from .constants import *

parse_date = make_date_mapper('%Y-%m-%d')


class AcisIO(JsonNetIO):
    """
    Base class for loading data from ACIS web services
    See http://data.rcc-acis.org/doc/
    """

    path = None  # ACIS web service path
    basin = None  # HUC8 watershed
    elems = None  # Requested element codes
    meta = None  # Requested metadata

    @property
    def url(self):
        """
        URL for wq.io.loaders.NetLoader
        """
        return "http://data.rcc-acis.org/%s" % self.path

    def getdate(self, name):
        """
        Retrieve given property from class/instance, ensuring it is a date
        """
        value = getattr(self, name, None)

        if value is None:
            return None

        if isinstance(value, basestring):
            value = parse_date(value)

        if isinstance(value, datetime):
            value = value.date()

        return value

    def getlist(self, name):
        """
        Retrieve given property from class/instance, ensuring it is a list.
        Also determine whether the list contains simple text/numeric values or
        nested dictionaries (a "complex" list)
        """
        value = getattr(self, name, None)
        complex = False

        def str_value(val):
            if isinstance(val, dict):
                complex = True
                return val
            else:
                return unicode(val)

        if value is None:
            pass
        elif isinstance(value, list):
            value = [str_value(val) for val in value]
        else:
            value = [str_value(value)]

        return value, complex

    def set_param(self, into, name):
        """
        Set parameter key, noting whether list value is "complex"
        """
        value, complex = self.getlist(name)
        if value is not None:
            into[name] = value
        return complex

    def get_params(self):
        """
        Get parameters for web service, noting whether any are "complex"
        """
        params = {}
        complex = False

        if self.basin is None:
            # FIXME: Support other region types
            raise NotImplementedError("No region specified")

        for param in ('basin', 'meta', 'elems'):
            if self.set_param(params, param):
                complex = True
        for param in ('sdate', 'edate'):
            value = self.getdate(param)
            if value:
                params[param] = [unicode(value)]
        return params, complex

    @property
    def params(self):
        """
        URL parameters for wq.io.loaders.NetLoader
        """
        params, complex = self.get_params()
        if complex:
            # ACIS web service supports JSON object as "params" parameter
            params = {
                key: val[0]
                if len(val) == 1 and isinstance(val[0], basestring)
                else val
                for key, val in params.items()
            }
            return {'params': json.dumps(params)}
        else:
            # Simpler queries can use traditional URL parameters
            return {
                key: ','.join(val)
                for key, val in params.items()
            }


class StationMetaIO(AcisIO):
    """
    Retrieves metadata about the climate stations in a region.
    See http://data.rcc-acis.org/doc/#title8
    """

    namespace = "meta"  # For wq.io.parsers.text.JsonParser
    path = "StnMeta"
    elems = None

    @property
    def meta(self):
        """
        Request all available metadata by default
        """
        meta = ['name', 'state', 'sids', 'll',
                'elev', 'uid', 'county', 'climdiv']
        if self.elems is not None:
            meta.append('valid_daterange')
        return meta

    def map_value(self, field, value):
        """
        Clean up some values returned from the web service.
        (overrides wq.io.mappers.BaseMapper)
        """

        if field == 'sids':
            # Site identifiers are returned as "[id] [auth_id]";
            # Map to auth name for easier usability
            ids = {}
            for idinfo in value:
                id, auth = idinfo.split(' ')
                auth = AUTHORITY_BY_ID[auth]
                ids[auth['name']] = id
            return ids

        if field == 'valid_daterange':
            # Date ranges for each element are returned in an array
            # (sorted by the order the elements were were requested);
            # Convert to dictionary with element id as key
            elems, complex = self.getlist('elems')
            ranges = {}
            for elem, val in zip(elems, value):
                if val:
                    start, end = val
                    ranges[elem] = (parse_date(start), parse_date(end))
                else:
                    ranges[elem] = None, None
            return ranges
        return value


class StationDataIO(StationMetaIO):
    """
    Retrieve daily time series data from the climate stations in a region.
    See http://data.rcc-acis.org/doc/#title19
    """

    namespace = "data"  # For wq.io.parsers.text.JsonParser
    path = "MultiStnData"

    sdate = None  # Start date
    edate = None  # End date
    add = None  # Additional information for daily results

    def get_field_names(self):
        """
        ACIS web service returns "meta" and "data" for each station;
        Use meta attributes as field names
        """
        field_names = super(StationDataIO, self).get_field_names()
        if field_names == ['meta', 'data']:
            field_names = self.data[0]['meta'].keys()
        return field_names

    def get_params(self):
        if self.sdate is None or self.edate is None:
            raise NotImplementedError("sdate and edate must be set!")
        params, complex = super(StationDataIO, self).get_params()

        # If self.add is defined, set it for each requested element
        if self.add is not None:
            add, add_is_complex = self.getlist('add')
            complex = True
            elems = []
            for elem in params.get('elems', []):
                if not isinstance(elem, dict):
                    elem = {'name': elem}
                elem['add'] = ",".join(add)
                elems.append(elem)
            params['elems'] = elems

        return params, complex

    def usable_item(self, data):
        """
        ACIS web service returns "meta" and "data" for each station; use meta
        attributes as item values, and add an IO for iterating over "data"
        """

        # Use metadata as item
        item = data['meta']

        # Add nested IO for data
        elems, elems_is_complex = self.getlist('elems')
        if elems_is_complex:
            elems = [elem['name'] for elem in elems]

        add, add_is_complex = self.getlist('add')
        item['data'] = DataIO(
            data=data['data'],
            elems=elems,
            add=add,
            sdate=self.getdate('sdate'),
            edate=self.getdate('edate')
        )

        # TupleMapper will convert item to namedtuple
        return super(StationDataIO, self).usable_item(item)


class DataIO(TimeSeriesMapper, BaseIO):
    """
    IO for iterating over ACIS time series data.
    Created internally by StationDataIO; not meant to be used directly.
    """

    # Inherited from parent
    elems = []
    add = []
    sdate = None
    edate = None

    date_formats = []  # For TimeSeriesMapper

    def load_data(self, data):
        """
        MultiStnData data results are arrays without explicit dates;
        Infer time series based on start date.
        """

        date = self.sdate
        day = timedelta(days=1)
        for row in data:
            data = {'date': date}
            if self.add:
                # If self.add is set, results will contain additional
                # attributes (e.g. flags). In that case, create one row per
                # result, with attributes "date", "elem", "value", and one for
                # each item in self.add.
                for elem, vals in zip(self.elems, row):
                    data['elem'] = elem
                    for add, val in zip(['value'] + self.add, vals):
                        data[add] = val
                    yield data
            else:
                # Otherwise, return one row per date, with "date" and each
                # element's value as attributes.
                for elem, val in zip(self.elems, row):
                    # namedtuple doesn't like numeric field names
                    if elem.isnumeric():
                        elem = "e%s" % elem
                    data[elem] = val
                yield data
            date += day

    def __init__(self, *args, **kwargs):
        data = kwargs.pop('data')
        self.data = self.load_data(data)
        super(DataIO, self).__init__(*args, **kwargs)

    def get_field_names(self):
        """
        Different field names depending on self.add setting (see load_data)
        For BaseIO
        """
        if self.add:
            return ['date', 'elem', 'value'] + [flag for flag in self.add]
        else:
            field_names = ['date']
            for elem in self.elems:
                # namedtuple doesn't like numeric field names
                if elem.isnumeric():
                    elem = "e%s" % elem
                field_names.append(elem)
            return field_names

    @property
    def key_fields(self):
        """
        Different key fields depending on self.add setting (see load_data)
        For TimeSeriesMapper
        """
        if self.add:
            return ['date', 'elem']
        else:
            return ['date']



#!/usr/bin/env python
import sys
from datetime import date
from climata import StationMetaIO, ELEMENT_BY_NAME, ELEMENT_BY_ID

elems = ELEMENT_BY_NAME.copy()

# Eloement 7 (pan evap) does not have a name, copy from ID listing
elems['7'] = ELEMENT_BY_ID['7']


def load_sites(*basin_ids):
    """
    Load metadata for all sites in given basin codes.
    """

    # Resolve basin ids to HUC8s if needed
    basins = []
    for basin in basin_ids:
        if basin.isdigit() and len(basin) == 8:
            basins.push(basin)
        else:
            from climata.huc8 import get_huc8
            basins.extend(get_huc8(basin))

    # Load sites with data since 1900
    sites = StationMetaIO(
        basin=basins,
        elems=elems.keys(),
        sdate='1900-01-01',
        edate=date.today(),
    )

    # Load all sites (to get sites without data)
    seen_sites = [site.uid for site in sites]
    nodata_sites = [
        site for site in StationMetaIO(basin=basins)
        if site.uid not in seen_sites
    ]

    # Determine the following from the site lists:
    seen_auths = set()  # Which authority codes are actually used by any site
    seen_elems = set()  # Which elems actually have data in any site
    ranges = {}  # The overall period of record for each site
    for site in sites:

        for auth in site.sids.keys():
            seen_auths.add(auth)

        start, end = None, None
        for elem in site.valid_daterange:
            s, e = site.valid_daterange[elem]
            seen_elems.add(elem)
            if s is None or e is None:
                continue
            if start is None or s < start:
                start = s
            if end is None or e > end:
                end = e
        ranges[site.uid] = [start, end]

    # Check for authority codes that might not be in sites with data
    for site in nodata_sites:
        for auth in site.sids.keys():
            seen_auths.add(auth)

    # Print CSV headers (FIXME: use CsvFileIO for this?)
    seen_auths = sorted(seen_auths)
    seen_elems = sorted(seen_elems)
    print ",".join(
        ['name']
        + seen_auths
        + ['latitude', 'longitude', 'start', 'end', 'years']
        + [elems[elem]['desc'] for elem in seen_elems]
    )

    # Print sites with data
    for site in sites:

        # Determine if elems are available for entire period or shorter range
        start, end = ranges[site.uid]
        if start and end:
            years = end.year - start.year + 1
        elem_ranges = []
        for elem in seen_elems:
            estart, eend = site.valid_daterange[elem]
            if estart is None:
                erange = ""
            elif estart == start and eend == end:
                erange = "period"
            else:
                erange = "%s to %s" % (estart.date(), eend.date())
            elem_ranges.append(erange)

        # Output CSV row
        print ",".join(map(str,
            [site.name]
            + [site.sids.get(auth, "") for auth in seen_auths]
            + [site.ll[1], site.ll[0]]
            + [start.date(), end.date(), years]
            + elem_ranges
        ))

    # Print CSV rows for sites without data
    for site in nodata_sites:
        print ",".join(
            [site.name]
            + [site.sids.get(auth, "") for auth in seen_auths]
            + [str(site.ll[1]), str(site.ll[0])]
            + ["NO DATA"]
        )

if __name__ == "__main__":
    if len(sys.argv) < 2:
        print "Usage: acis_sites.py basin"
        exit()
    load_sites(*sys.argv[1:])



from wq.io import CsvNetIO


class HucIO(CsvNetIO):
    # FIXME: some HUCs have changed, but new_huc_rdb.txt isn't a valid TSV file
    url = "http://water.usgs.gov/GIS/huc_rdb.txt"
    delimiter = "\t"

hucs = list(HucIO())


def get_huc8(prefix):
    """
    Return all HUC8s matching the given prefix (e.g. 1801) or basin name
    (e.g. Klamath)
    """
    if not prefix.isdigit():
        # Look up hucs by name
        name = prefix
        prefix = None
        for row in hucs:
            if row.basin.lower() == name.lower():
                # Use most general huc if two have the same name
                if prefix is None or len(row.huc) < len(prefix):
                    prefix = row.huc

    if prefix is None:
        return []

    huc8s = []
    for row in hucs:
        # Return all 8-digit hucs with given prefix
        if len(row.huc) == 8 and row.huc.startswith(prefix):
            huc8s.append(row.huc)
    return huc8s



#!/usr/bin/env python
from climata import StationMetaIO, StationDataIO, ELEMENT_BY_NAME
from climata.io import parse_date
import sys
from datetime import date, timedelta

curyear = date.today().year
one_day = timedelta(days=1)


def load_data(basin, elem, syear=1950, eyear=curyear,
              inactive=False, years=30):
    syear = int(syear)
    eyear = int(eyear)
    inactive = bool(inactive)
    years = int(years)
    if not basin.isdigit() or len(basin) != 8:
        from climata.huc8 import get_huc8
        basin = get_huc8(basin)

    sites = StationMetaIO(
        basin=basin,
        sdate='%s-01-01' % syear,
        edate='%s-12-31' % eyear,
        elems=elem,
    )
    include_sites = []
    seen_auths = set()
    for site in sites:
        start, end = site.valid_daterange[elem]
        exclude = False
        if end.year < eyear - years and not inactive:
            sys.stderr.write("Site not active: %s (last %s)" % (
                site.name, end.date()
            ))
            sys.stderr.write("\n")
            exclude = True
        elif end.year - start.year + 1 < years:
            sys.stderr.write("Not enough data: %s (%s years)" % (
                site.name, end.year - start.year
            ))
            sys.stderr.write("\n")
            exclude = True

        if exclude:
            continue
        include_sites.append(site)
        for auth in site.sids.keys():
            seen_auths.add(auth)

    # Sort sites by longitude
    include_sites = sorted(include_sites, key=lambda s: s.ll[0])
    seen_auths = sorted(seen_auths)

    def get_val(site, field):
        if hasattr(site, field):
            return getattr(site, field)
        elif field == "latitude":
            return site.ll[1]
        elif field == "longitude":
            return site.ll[0]
        else:
            return site.sids.get(field, "")

    def header(field):
        vals = [get_val(site, field) for site in include_sites]
        sys.stdout.write(",".join([field + ":"] + map(str, vals)))
        sys.stdout.write("\n")

    header("name")
    for auth in seen_auths:
        header(auth)
    header("latitude")
    header("longitude")

    for year in range(syear, eyear + 1):
        load_year_data(basin, elem, year, include_sites)


def load_year_data(basin, elem, year, include_sites):
    sys.stderr.write("Loading %s data...\n" % year)
    sdate = parse_date('%s-01-01' % year).date()
    edate = parse_date('%s-12-31' % year).date()
    sitedata = StationDataIO(
        basin=basin,
        sdate=sdate,
        edate=edate,
        elems=elem,
    )
    sitedata = {site.uid: site for site in sitedata}
    dates = {}
    date = sdate
    while date <= edate:
        dates[date] = {}
        date += one_day

    for site in include_sites:
        if site.uid not in sitedata:
            continue
        data = sitedata[site.uid].data
        for row in data:
            dates[row.date][site.uid] = getattr(row, elem)

    date = sdate
    while date <= edate:
        data = map(str, [
            dates[date].get(site.uid, "")
            for site in include_sites
        ])
        sys.stdout.write(",".join([str(date)] + data))
        sys.stdout.write("\n")
        date += one_day


def usage():
    sys.stderr.write("""
Usage: acis_data.py basin elem [syear] [eyear] [inact] [years]

 basin:\tbasin(s) (HUC8, required)
 elem:\telement code (required, see below)
 syear:\tStart year (default 1950)
 eyear:\tEnd year (default %s)
 inact:\tInclude inactive sites (default false)
 years:\tOnly include sites with this many years of data (default 30)

Inactive sites are considered to be sites that have not had any data for n
years, where n is the same as the years argument.

Available elem codes:
""" % curyear)

    for elem in sorted(ELEMENT_BY_NAME.keys()):
        sys.stderr.write(" %s:\t%s\n" % (elem, ELEMENT_BY_NAME[elem]['desc']))
    exit()

if __name__ == "__main__":
    if len(sys.argv) < 3 or sys.argv[2] not in ELEMENT_BY_NAME:
        usage()
        exit()
    load_data(*sys.argv[1:])



import os
from setuptools import setup, find_packages

LONG_DESCRIPTION = """
A python library for iterating over Applied Climate Information System (ACIS)
time series data.  Powered by wq.io.
"""

def long_description():
    """Return long description from README.rst if it's present
    because it doesn't get installed."""
    try:
        return open(os.path.join(os.path.dirname(__file__), 'README.rst')).read()
    except IOError:
        return LONG_DESCRIPTION


setup(
    name='climata',
    version='0.1.0',
    author='HEI Geo',
    author_email='gis@houstoneng.com',
    url='https://github.com/heigeo/climata',
    license='MIT',
    packages=find_packages(),
    description='A pythonic interface for Applied Climate Information System (ACIS) data',
    long_description=long_description(),
    install_requires=['wq.io>=0.3.0'],
    scripts=['climata/bin/acis_sites.py', 'climata/bin/acis_data.py'],
    classifiers = [
        'Development Status :: 2 - Pre-Alpha',
        'Environment :: Web Environment',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
    ]
)
